# nw-filter.awk -- a noweave filter adapting noweb code chunks
#                  containing Python or Ada to some Knuthian notation
#
# Some adaptions are done here:
#   - comment lines surrounded by empty lines are replaced by a
#     LaTeX \separationRule (to be defined in the noweb file);
#   - all code chunks are surrounded either by an interfaceModule or
#     otherModule LaTeX environment (to be defined in the noweb file)
#     depending on the chunk name;
#   - all document chunks have empty lines at their beginning to
#     enable page breaking before the chunk (this is CWEB style)
#
# Author: Dr. Thomas Tensi, 2013

BEGIN \
  {
      true = 1;
      false = 0;

      inCodeChunk = false;
      bufferLength = 0;

      separationRuleCommand = "@literal \\separationRule";

      # here we configure whether the code chunks shall be
      # bracketed with a special environment
      codeChunksAreBracketed = true;

      # those are the modules without an explicit interface
      moduleList = ("base|boolean|boundedscalar|character" \
		    "|conditionalexpression|engine" \
		    "|genericconditionalexpression|generictrace" \
		    "|homogenous one-d map|integer|longnatural|one-d map" \
                    "|presentation|set|static allocator|toplevel");

      # the following variable tells whether the interface module
      # brackets go around the complete code chunk (or only around the
      # code itself)
      moduleBracketStartsAtChunkBegin = false;

      moduleLeadinCommand["interface"]  = "@literal \\begin{interfaceModule}";
      moduleLeadinCommand["other"]      = "@literal \\begin{otherModule}";
      moduleLeadoutCommand["interface"] = "@literal \\end{interfaceModule}";
      moduleLeadoutCommand["other"]     = "@literal \\end{otherModule}";

      # we replace a blank line followed by a separator follows by
      # a blank line with a separation rule
      State_atModuleBegin        = 0;
      State_inLimbo              = 1;
      State_inBlankText          = 2;
      State_afterBlankTextLine   = 3;
      State_afterRuleText        = 4;
      State_afterRuleLine        = 5;
      State_inBlankTextAfterRule = 6;
  }

#--------------------

/.*/ \
  {
      currentLine = $0;
  }

#--------------------

/@defn/ \
  {
      gsub("_", "\\_", currentLine);

      if (!codeChunksAreBracketed) {
          isInterfaceModule = false;
      } else {
          moduleName = substr(currentLine, 7, 1000);
          isInterfaceModule = isInterfaceModuleName(moduleName);
	  moduleKind = (isInterfaceModule ? "interface" : "other");

          if (!moduleBracketStartsAtChunkBegin) {
              state = State_atModuleBegin;
          }
      }
  }

#--------------------

/(@use|@xref)/ \
  {
      gsub("_", "\\_", currentLine);
  }

#--------------------

/^@end code/ \
  {
      if (codeChunksAreBracketed) {
          if (moduleBracketStartsAtChunkBegin) {
              print moduleLeadinCommand[moduleKind];
          } else {
              appendToBuffer(moduleLeadoutCommand[moduleKind]);
          }
      }

      appendToBuffer(currentLine);

      if (codeChunksAreBracketed && moduleBracketStartsAtChunkBegin) {
          appendToBuffer(moduleLeadoutCommand[moduleKind]);
      }

      flushBuffer();
      inCodeChunk = false;
      next;
  }

#--------------------

/^@begin code/ \
  {
      inCodeChunk = true;
      isInterfaceModule = false;
      state = State_inLimbo;
      flushBuffer();
      appendToBuffer(currentLine);
      next;
  }

#--------------------

/^@begin docs/ \
  {
      print currentLine;
      # insert dummy blank line to enable page breaking before by
      # noweb conventions
      printf "%s", "@text \n@nl\n@text \n@nl\n";
      next;
  }

#--------------------

/.*/ && !inCodeChunk \
  {
      print currentLine;
      next;
  }

#--------------------

inCodeChunk \
  {
      if (currentLine == "@nl") {
          if (state == State_atModuleBegin) {
              appendToBuffer(currentLine);
              currentLine = moduleLeadinCommand[moduleKind];
              state = State_inLimbo;
          } else if (state == State_inBlankText \
              || state == State_afterRuleText) {
              state++;
          } else if (state != State_inBlankTextAfterRule) {
              state = State_inLimbo;
          } else {
              # retain text and separation rule
              textPart = buffer[bufferLength - 3];
              literalPart = buffer[bufferLength - 2];
              bufferLength -= 6;
              state = State_inLimbo;
              appendToBuffer(textPart);
              appendToBuffer(literalPart);
          }
      } else if (currentLine !~ /^@text/) {
          if (state > State_inLimbo) {
              state = State_inLimbo;
          }
      } else {
          effectiveText = substr(currentLine, 7, 1000);
          
          if (effectiveText == "") {
              if (state == State_afterBlankTextLine) {
                  state = State_inBlankText;
              } else if (state == State_inLimbo \
                  || state == State_afterRuleLine) {
                  state++;
              }
          } else if (effectiveText ~ /\-\- \-\-\-/ ) {
              if (state == State_afterBlankTextLine) {
                  state++;
              }

              sub(/\-\-.*/, "", currentLine);
              appendToBuffer(currentLine);
              currentLine = separationRuleCommand;
          } else {
              # do some cosmetic replacements
              gsub(/\/=/, "<>", effectiveText); # not equal
              currentLine = "@text " effectiveText;
              state = State_inLimbo;
          }
      }

      appendToBuffer(currentLine);
      next;
  }

#--------------------

END \
  {
      flushBuffer();
  }

#--------------------
#--------------------

function appendToBuffer (st) \
  {
      # appends <st> to buffer
      ## print "append in " state " @" bufferLength+1 ": " st
      buffer[++bufferLength] = st;
  }

#--------------------

function expandToLiteralMarkup (st) \
  {
      # makes "@literal" markup from <st>
      partCount = split(st, partList, "\n");
      result = "";

      for (i = 1;  i <= partCount;  i++) {
          result = result "@literal " partList[i] "\n@nl";
          result = result (i < partCount? "\n" : "");
      }

      return result;
  }

#--------------------

function flushBuffer () \
  {
      # remove trailing blank lines
      for (i = bufferLength;  i > 0;  i--) {
          thisLine = buffer[i];
          
          if (thisLine ~ /^@text *$/ || thisLine ~ /^@nl$/ ) {
              bufferLength--;
          } else {
              break;
          }
      }
    
      for (i = 1;  i <= bufferLength;  i++) {
          print buffer[i];
      }
    
      bufferLength = 0;
  }

#--------------------

function isInterfaceModuleName (moduleName) \
  {
      # checks whether <moduleName> denotes an interface module code
      # chunk

      moduleListPattern = "^(" moduleList ") module specification$";
      result = (moduleName ~ moduleListPattern \
                || moduleName ~ /module interface$/);
      return result;
  }

#--------------------

function stripPreceedingNl () \
  {
      # removes a previous line when it is just a @nl
      if (bufferLength > 0 && buffer[bufferLength] ~ /^@nl/ \
          && buffer[bufferLength - 1] == "@text ") {
          bufferLength -= 2;
      }
  }
